from aegisx.ext.jose import TokenValidator
from aegisx.ext.jose.models import JWSHeader
from libcanonical.types import HTTPResourceLocator

from aegisx.ext.oauth.models import OIDCToken
from aegisx.ext.oauth.models import ServerMetadata


class OIDCTokenValidator(TokenValidator[OIDCToken]):

    def __init__(
        self,
        audience: set[str] | None = None,
        issuer: set[str] | None = None
    ):
        super().__init__(
            types=OIDCToken,
            audience=audience,
            issuer=issuer
        )

    async def get_verification_jwks(
        self,
        header: JWSHeader,
        payload: OIDCToken
    ):
        jwks = self.jwks
        if isinstance(payload.iss, HTTPResourceLocator)\
        and self.is_trusted_issuer(payload.iss):
            metadata = await ServerMetadata.get(payload.iss)
            jwks = metadata.jwks
        return jwks


from typing import Literal


AuthenticationMethodReferenceLiteral = Literal[
    'face',
    'fpt',
    'geo',
    'hwk',
    'iris',
    'kba',
    'mca',
    'mfa',
    'otp',
    'pin',
    'pop',
    'pwd',
    'rba',
    'retina',
    'sc',
    'sms',
    'swk',
    'tel',
    'user',
    'vbm',
    'wia',
]


import enum


class ClientAuthenticationMethod(str, enum.Enum):
    none                        = 'none'
    client_secret_post          = 'client_secret_post'
    client_secret_basic         = 'client_secret_basic'
    client_secret_jwt           = 'client_secret_jwt'
    private_key_jwt             = 'private_key_jwt'
    tls_client_auth             = 'tls_client_auth'
    self_signed_tls_client_auth = 'self_signed_tls_client_auth'


import enum


class ResponseType(str, enum.Enum):
    CODE                = 'code'
    CODE_ID_TOKEN       = 'code id_token'
    CODE_ID_TOKEN_TOKEN = 'code id_token token'
    CODE_TOKEN          = 'code token'
    ID_TOKEN_TOKEN      = 'id_token token'
    ID_TOKEN            = 'id_token'
    TOKEN               = 'token'
    NONE                = 'none'


from typing import cast
from typing import Any
from typing import Generic
from typing import TypeVar

import pydantic


I = TypeVar('I', bound=Any)
T = TypeVar('T')


class ClaimRequest(pydantic.BaseModel, Generic[T]):
    essential: bool = pydantic.Field(
        default=False,
        title="Essential",
        description=(
            "Indicates whether the claim being requested is essential. "
            "If the value is `true`, this indicates that the claim is an "
            "essential claim."
        )
    )

    values: list[T] | None = pydantic.Field(
        default=None,
        title="Values",
        min_length=1,
        description=(
            "Requests that the claim be returned with one of a set of values, "
            "with the values appearing in order of preference. This is "
            "processed equivalently to a value request, except that a choice "
            "of acceptable claim values is provided."
        )
    )

    @pydantic.model_validator(mode='before')
    @classmethod
    def preprocess(cls, value: I) -> I:
        if isinstance(value, dict):
            value = cast(dict[str, Any], value) # type: ignore
            if value.get('value') and value.get('values'):
                raise ValueError(
                    'the "value" and "values" parameters '
                    'are mutually exclusive.'
                )
            if 'value' in value:
                value['values'] = [value.pop('value')]
        return value

    @pydantic.model_serializer(mode='plain', when_used='json')
    def serialize(self) -> None | dict[str, bool | list[T]]:
        adapter: pydantic.TypeAdapter[list[T]] = pydantic.TypeAdapter(ClaimRequest.model_fields['values'].annotation)
        return None if self.is_voluntary() else {
            'essential': self.essential,
            'values': adapter.dump_python(self.values) # type: ignore
        }

    def is_voluntary(self) -> bool:
        return self.values is None and not self.essential


import pydantic

from ._accesstokentokenresponse import AccessTokenTokenResponse


class ProofOfPossessionTokenResponse(AccessTokenTokenResponse):
    cnf: str | None = pydantic.Field(
        default=None,
        description=(
            "Present if the token type is `pop` and a symmetric key is "
            "used. MAY be present for asymmetric PoP keys. This field "
            "contains the PoP key that the AS selected for the token. "
            "Values of this parameter follow the syntax and semantics "
            "of the `cnf` claim either from Section 3.1 of RFC 8747 for "
            "CBOR-based interactions or from Section 3.1 of RFC 7800 for "
            "JSON-based interactions."
        )
    )

    rs_cnf: str | None = pydantic.Field(
        default=None,
        description=(
            "Present if the token type is `pop` and asymmetric keys are used. "
            "Not present otherwise. This field contains information about the "
            "public key used by the RS to authenticate. If this parameter is "
            "absent, either the RS does not use a public key or the AS knows "
            "that the RS can authenticate itself to the client without additional "
            "information. Values of this parameter follow the syntax and semantics "
            "of the cnf claim either from Section 3.1 of RFC 8747 for CBOR-based "
            "interactions or from Section 3.1 of RFC 7800 for JSON-based interactions."
        )
    )

    @pydantic.model_validator(mode='after')
    def postprocess(self):
        if not self.cnf and not self.rs_cnf:
            raise ValueError("not a proof-of-possession token response.")
        return self


from typing import Literal
from typing import TypeAlias


GrantTypeLiteral: TypeAlias = Literal[
    'authorization_code',
    'client_credentials',
    'refresh_token',
    'urn:ietf:params:oauth:grant-type:jwt-bearer',
    'urn:ietf:params:oauth:grant-type:saml2-bearer',
    'urn:ietf:params:oauth:grant-type:token-exchange',
    'urn:ietf:params:oauth:grant-type:device_code',
    'urn:ietf:params:oauth:grant-type:token-exchange'
]



from typing import TYPE_CHECKING

import pydantic
from aegisx.ext.jose.types import JWECompactEncoded
from aegisx.ext.jose.types import JWSCompactEncoded

from ._authorizationserverresponse import AuthorizationServerResponse
from ._fields import AccessTokenExpiresInField
from ._fields import AccessTokenField
from ._fields import AccessTokenScopeField
from ._fields import AccessTokenTypeField
from ._fields import StateField
from ._oidctoken import OIDCToken
from ._tokenresponsemixin import TokenResponseMixin
if TYPE_CHECKING:
    from aegisx.ext.oauth.client import OIDCTokenValidator


class AccessTokenTokenResponse(AuthorizationServerResponse, TokenResponseMixin):
    model_config = {'extra': 'ignore'}

    access_token: AccessTokenField
    token_type: AccessTokenTypeField
    expires_in: AccessTokenExpiresInField
    scope: AccessTokenScopeField
    state: StateField

    refresh_token: str | None = pydantic.Field(
        default=None,
        title="Refresh token",
        description=(
            "The refresh token, which can be used to obtain new access tokens "
            "using the same authorization grant as described in Section 6 of "
            "RFC 6749."
        )
    )

    id_token: JWECompactEncoded | JWSCompactEncoded | None = pydantic.Field(
        default=None,
        title="ID Token",
        description=(
            "JSON Web Token (JWT) that contains claims about the authentication "
            "event. It MAY contain other claims."
        )
    )

    jwt: OIDCToken | None = pydantic.Field(
        default=None,
        exclude=True
    )

    async def validate_id_token(
        self,
        validator: 'OIDCTokenValidator'
    ):
        self.jwt = await validator.validate(self.id_token)


import pydantic
from libcanonical.utils import deephash

from aegisx.ext.oauth.types import HTTPSResourceLocator
from aegisx.ext.oauth.types import IssuerIdentifier
from ._servermetadata import ServerMetadata


class ClientConfiguration(pydantic.BaseModel):
    """Configuration parameters for OAuth 2.x/OpenID Connect clients."""
    model_config = {'extra': 'forbid'}

    issuer: IssuerIdentifier = pydantic.Field(
        default=...
    )

    client_id: str = pydantic.Field(
        default=...
    )

    client_secret: str | None = pydantic.Field(
        default=None
    )

    authorization_endpoint: HTTPSResourceLocator | None = pydantic.Field(
        default=None
    )

    token_endpoint: HTTPSResourceLocator | None = pydantic.Field(
        default=None
    )

    metadata_url: HTTPSResourceLocator | None = pydantic.Field(
        default=None
    )

    default_redirect_uri: str | None = pydantic.Field(
        default=None
    )

    discoverable: bool = pydantic.Field(
        default=False
    )

    @property
    def metadata(self):
        metadata = ServerMetadata.model_validate({
            'issuer': self.issuer,
            'authorization_endpoint': self.authorization_endpoint,
            'token_endpoint': self.token_endpoint,
            'metadata_url': self.metadata_url,
        })
        if self.discoverable:
            metadata.discovered = False
        return metadata

    @property
    def __key__(self):
        return self.key(self.issuer, self.client_id)

    @staticmethod
    def key(issuer: IssuerIdentifier, client_id: str):
        return deephash((issuer, client_id), using='sha256', encode='hex')


import pydantic


class TokenResponseMixin:

    @pydantic.field_validator('token_type', mode='before')
    @classmethod
    def preprocess_token_type(cls, value: str | None):
        if value is not None:
            value = str.lower(value)
        return value


import pydantic
from aegisx.ext.jose.types import JWECompactEncoded
from aegisx.ext.jose.types import JWSCompactEncoded

from ._authorizationserverresponse import AuthorizationServerResponse


class JARMAuthorizationServerResponse(AuthorizationServerResponse):
    response: JWECompactEncoded | JWSCompactEncoded = pydantic.Field(
        default=...,
        title="Response",
        description=(
            "The JWT Secured Authorization Response (JAR). The JWT is either "
            "signed, or signed and encrypted. If the JWT is both signed and "
            "encrypted, the JSON document will be signed then encrypted, with "
            "the result being a Nested JWT, as defined in RFC 7519"
        )
    )

    def is_encrypted(self) -> bool:
        return isinstance(self.response, JWECompactEncoded)

    def is_signed(self) -> bool:
        return True


from typing import Literal

import pydantic

from ._tokenrequestbase import TokenRequestBase


class AuthorizationCodeGrant(TokenRequestBase[Literal['authorization_code']]):
    code: str = pydantic.Field(
        default=...,
        title="Code",
        description="The authorization code received from the authorization server."
    )

    redirect_uri: str | None = pydantic.Field(
        default=None,
        title="Redirect URI",
        description=(
            "Required if the `redirect_uri` parameter was included in the "
            "authorization request as described in Section 4.1.1 of RFC 6749, "
            "and their values MUST be identical."
        )
    )


import contextlib
import logging
import json
import os
from typing import Any
from typing import Union

import httpx
import pydantic
from aegisx.ext.jose import JSONWebKeySet

from aegisx.ext.oauth.types import ClientAuthenticationMethod
from aegisx.ext.oauth.types import HTTPSResourceLocator
from aegisx.ext.oauth.types import NotDiscoverable


logger: logging.Logger = logging.getLogger(__name__)

DISCOVERED: dict[str, 'ServerMetadata'] = {}


class ServerMetadata(pydantic.BaseModel):
    issuer: str = pydantic.Field(
        default=...,
        title="Issuer",
        description="Authorization server's issuer identifier URL.",
    )

    authorization_endpoint: HTTPSResourceLocator | None = pydantic.Field(
        default=None,
        title="Authorization endpoint",
        description="URL of the authorization server's authorization endpoint.",
    )

    token_endpoint: HTTPSResourceLocator | None = pydantic.Field(
        default=None,
        title="Token endpoint",
        description="URL of the authorization server's token endpoint.",
    )

    jwks_uri: HTTPSResourceLocator | None = pydantic.Field(
        default=None,
        title="JSON Web Key Set (JWKS) URI",
        description="URL of the authorization server's JWK Set document.",
    )

    registration_endpoint: HTTPSResourceLocator | None = pydantic.Field(
        default=None,
        title="Registration endpoint",
        description=(
            "URL of the authorization server's **OAuth 2.0 Dynamic Client "
            "Registration Endpoint**."
        ),
    )

    scopes_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported scopes",
        description=(
            "JSON array containing a list of the OAuth 2.0 `scope` values that "
            "this authorization server supports."
        ),
    )

    response_types_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported response types",
        description=(
            "JSON array containing a list of the OAuth 2.0 `response_type` "
            "values that this authorization server supports."
        ),
    )

    response_modes_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported response modes",
        description=(
            "JSON array containing a list of the OAuth 2.0 `response_mode` "
            "values that this authorization server supports."
        ),
    )

    grant_types_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported grant types",
        description=(
            "JSON array containing a list of the OAuth 2.0 `grant_types` "
            "values that this authorization server supports."
        ),
    )

    token_endpoint_auth_methods_supported: list[ClientAuthenticationMethod] = pydantic.Field(
        default=[],
        title="Supported client authentication methods",
        description=(
            "JSON array containing a list of client authentication methods "
            "supported by this token endpoint."
        ),
    )

    token_endpoint_auth_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the token endpoint for the signature on the JWT "
            "used to authenticate the client at the token endpoint."
        ),
    )

    service_documentation: str | None = pydantic.Field(
        default=None,
        title="Documentation",
        description=(
            "URL of a page containing human-readable information that "
            "developers might want or need to know when using the "
            "authorization server."
        ),
    )

    ui_locales_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported locals for UI",
        description=(
            "Languages and scripts supported for the user interface, "
            "represented as a JSON array of language tag values from BCP 47."
        ),
    )

    op_policy_uri: str | None= pydantic.Field(
        default=None,
        title="Data policy URL",
        description=(
            "URL that the authorization server provides to the person "
            "registering the client to read about the authorization server's "
            "requirements on how the client can use the data provided by the "
            "authorization server."
        ),
    )

    op_tos_uri: str | None= pydantic.Field(
        default=None,
        title="Terms of service URL",
        description=(
            "URL that the authorization server provides to the person "
            "registering the client to read about the authorization server's "
            "terms of service."
        ),
    )

    revocation_endpoint: HTTPSResourceLocator | None = pydantic.Field(
        default=None,
        title="Revocation endpoint",
        description="URL of the authorization server's revocation endpoint.",
    )

    revocation_endpoint_auth_methods_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported authentication methods",
        description=(
            "JSON array containing a list of client authentication methods "
            "supported by this revocation endpoint."
        ),
    )

    revocation_endpoint_auth_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the revocation endpoint for the signature on the JWT "
            "used to authenticate the client at the revocation endpoint."
        ),
    )

    introspection_endpoint: HTTPSResourceLocator | None = pydantic.Field(
        default=None,
        title="Introspection endpoint",
        description="URL of the authorization server's introspection endpoint.",
    )

    introspection_endpoint_auth_methods_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported authentication methods",
        description=(
            "JSON array containing a list of client authentication methods "
            "supported by this introspection endpoint."
        ),
    )

    introspection_endpoint_auth_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the introspection endpoint for the signature on the JWT "
            "used to authenticate the client at the introspection endpoint."
        ),
    )

    signed_metadata: str | None = pydantic.Field(
        default=None,
        title="Signed metadata",
        description=(
            "Signed JWT containing metadata values about the authorization "
            "server as claims."
        )
    )

    device_authorization_endpoint: HTTPSResourceLocator | None = pydantic.Field(
        default=None,
        title="Device authorization endpoint",
        description=(
            "URL of the authorization server's device authorization endpoint."
        ),
    )

    tls_client_certificate_bound_access_tokens: bool | None = pydantic.Field(
        default=True,
        title="Supports mTLS certificate-bound access tokens",
        description=(
            "Indicates authorization server support for mutual-TLS client "
            "certificate-bound access tokens."
        ),
    )

    mtls_endpoint_aliases: dict[str, str] | None = pydantic.Field(
        default={},
        title="Alternative mTLS endpoints",
        description=(
            "JSON object containing alternative authorization server "
            "endpoints, which a client intending to do mutual TLS will "
            "use in preference to the conventional endpoints."
        ),
    )

    nfv_token_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as NFV Token."
        ),
    )

    nfv_token_encryption_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "NFV Token."
        ),
    )

    nfv_token_encryption_enc_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "NFV Token."
        ),
    )

    userinfo_endpoint: str | None = pydantic.Field(
        default=None,
        title="UserInfo endpoint",
        description="URL of the authorization servers' UserInfo Endpoint.",
    )

    acr_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported ACR",
        description=(
            "JSON array containing a list of the Authentication Context Class "
            "References that this authorization server supports."
        ),
    )

    subject_types_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported subject types",
        description=(
            "JSON array containing a list of the Subject Identifier types that "
            "this authorization server supports"
        ),
    )

    id_token_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as ID Token."
        ),
    )

    id_token_encryption_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "ID Token."
        ),
    )

    id_token_encryption_enc_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "ID Token."
        ),
    )

    userinfo_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as UserInfo Endpoint."
        ),
    )

    userinfo_encryption_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "UserInfo Endpoint."
        ),
    )

    userinfo_encryption_enc_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "UserInfo Endpoint."
        ),
    )

    request_object_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as Request Object."
        ),
    )

    request_object_encryption_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "Request Object."
        ),
    )

    request_object_encryption_enc_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "Request Object."
        ),
    )

    display_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported display modes",
        description=(
            "JSON array containing a list of the `display` parameter values "
            "that the OpenID Provider supports."
        ),
    )

    claim_types_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported Claims Types",
        description=(
            "JSON array containing a list of the Claims Types "
            "that the OpenID Provider supports."
        )
    )

    claims_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported Claims Types",
        description=(
            "JSON array containing a list of the Claim Names of the Claims "
            "that the OpenID Provider MAY be able to supply values for."
        )
    )

    claims_locales_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported claim locales",
        description=(
            "Languages and scripts supported for values in Claims being "
            "returned, represented as a JSON array of BCP 47."
        ),
    )

    claims_parameter_supported: bool | None = pydantic.Field(
        default=False,
        title="Supports `claims` parameter?",
        description=(
            "Boolean value specifying whether the OP supports use of the "
            "`claims` parameter."
        ),
    )

    request_parameter_supported: bool | None = pydantic.Field(
        default=False,
        title="Supports `request` parameter?",
        description=(
            "Boolean value specifying whether the OP supports use of the "
            "`request` parameter."
        ),
    )

    request_uri_parameter_supported: bool | None = pydantic.Field(
        default=False,
        title="Supports `request_uri` parameter?",
        description=(
            "Boolean value specifying whether the OP supports use of the "
            "`request_uri` parameter."
        ),
    )

    require_request_uri_registration: bool | None = pydantic.Field(
        default=True,
        title="Requires pre-regiration?",
        description=(
            "Boolean value specifying whether the OP requires any `request_uri` "
            "values used to be pre-registered."
        ),
    )

    require_signed_request_object: bool | None = pydantic.Field(
        default=True,
        title="Requires pre-regiration?",
        description=(
            "Indicates where authorization request needs to be protected as "
            "**Request Object** and provided through either `request` or "
            "`request_uri` parameter."
        ),
    )

    pushed_authorization_request_endpoint: HTTPSResourceLocator | None = pydantic.Field(
        default=None,
        title="Pushed Authorization Request (PAR) endpoint",
        description=(
            "URL of the authorization server's pushed authorization request "
            "endpoint."
        ),
    )

    require_pushed_authorization_requests: bool | None = pydantic.Field(
        default=False,
        title="Requires PAR?",
        description=(
            "Indicates whether the authorization server accepts authorization "
            "requests only via PAR."
        ),
    )

    introspection_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of algorithms supported by the "
            "authorization server for introspection response signing."
        ),
    )

    introspection_encryption_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of algorithms supported by the "
            "authorization server for introspection response content key "
            "encryption (`alg` value)."
        ),
    )

    introspection_encryption_enc_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of algorithms supported by the "
            "authorization server for introspection response content "
            "encryption (`enc` value)."
        ),
    )

    authorization_response_iss_parameter_supported: bool | None = pydantic.Field(
        default=False,
        title="Supports `iss` parameter in authorization response?",
        description=(
            "Boolean value indicating whether the authorization server "
            "provides the `iss` parameter in the authorization response."
        ),
    )

    authorization_signing_alg_values_supported: list[str] = pydantic.Field(
        default=[],
        title="JARM signature algorithms",
        description=(
            "A JSON array containing a list of the JWS signing algorithms "
            "(`alg` values) supported by the authorization endpoint to "
            "sign the response."
        )
    )

    authorization_encryption_alg_values_supported: list[str] = pydantic.Field(
        default=[],
        title="JARM key wrapping or key agreement algorithms",
        description=(
            "A JSON array containing a list of the JWS signing algorithms "
            "(`alg` values) supported by the authorization endpoint to "
            "encrypt the response."
        )
    )

    authorization_encryption_enc_values_supported: list[str] = pydantic.Field(
        default=[],
        title="JARM encryption algorithms",
        description=(
            "A JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the authorization endpoint to "
            "encrypt the response."
        )
    )

    signed_metadata: str | None = pydantic.Field(
        default=None,
        title="Signed metadata",
        description=(
            "A JWT containing metadata values about the authorization server as claims. "
            "This is a string value consisting of the entire signed JWT.  A `signed_metadata` "
            "metadata value SHOULD NOT appear as a claim in the JWT."
        )
    )

    jwks: JSONWebKeySet = pydantic.Field(
        default_factory=JSONWebKeySet,
        exclude=True
    )

    discovered: bool = pydantic.Field(
        default=False,
        exclude=True
    )

    metadata_url: str | None = pydantic.Field(
        default=None,
        exclude=True
    )

    @pydantic.model_validator(mode='before')
    def preprocess(cls, values: Union[dict[str, Any], 'ServerMetadata']):
        if isinstance(values, dict)\
        and values.get('issuer')\
        and DISCOVERED.get(values['issuer']):
            values = DISCOVERED[ values['issuer'] ].model_dump()
        return values

    @staticmethod
    def clear():
        """Clear the in-memory issuer cache."""
        DISCOVERED.clear()

    @staticmethod
    def is_cached(issuer: str):
        """Return a boolean indicating if the issuer is cached."""
        return bool(DISCOVERED.get(issuer))

    @classmethod
    async def get(cls, issuer: str):
        self = cls(issuer=issuer)
        await self.discover()
        return self

    def is_discovered(self):
        """Return a boolean indicating is the metadata is discovered."""
        return self.discovered

    async def discover(
        self,
        client: httpx.AsyncClient | None = None,
        metadata_url: str | None = None,
        force: bool = False
    ) -> bool:
        """Discover the authorization server metadata.

        By default, :meth:`ServerMetadata.discover()` uses the :attr:`issuer`
        to construct the URI of the metadata endpoint. If the server uses a
        non-standard or non-default metadata endpoint, it may be provided
        using the `metadata_url` parameter.

        The response from the authorization server will not override any
        properties already set on :class:`ServerMetadata`. This is to allow
        the implementer to provide their own values in cases where this is
        needed.

        Set the `force` parameter to ``True`` to force discovery for
        :class:`ServerMetadata` instances that have previously invoked
        the :meth:`discover()` method.
        """
        if self.is_discovered() and not force:
            return False
        metadata_url = self.metadata_url or metadata_url
        metadata = DISCOVERED.get(self.issuer)
        if metadata is None:
            async with self._client(client) as client:
                logger.debug("Retrieving server metadata (issuer: %s)", self.issuer)
                response = await self._discover_metadata_endpoint(client, metadata_url=metadata_url)
                if response is None:
                    raise NotDiscoverable(f"Unable to discover server metadata for {self.issuer}")
                try:
                    data = dict(response.json())
                    metadata = ServerMetadata.model_validate(data)
                except (TypeError, ValueError, json.JSONDecodeError, pydantic.ValidationError):
                    raise ValueError(
                        f"The metadata endpoint of {self.issuer} returned an unexpected "
                        f"response (url: {response.url})"
                    )

                # Fetch the JWKS. Do this before assigning the attributes so any failure
                # leaves the original fields untouched.
                jwks_uri = self.jwks_uri or metadata.jwks_uri
                if jwks_uri:
                    response = await client.get(url=jwks_uri)
                    if response.status_code != 200:
                        raise ValueError(
                            f"Unable to retrieve JWKS for {self.issuer}"
                            f": {response.status_code}"
                        )
                    try:
                        data = dict(response.json())
                        self.jwks = JSONWebKeySet.model_validate(data)
                    except (TypeError, ValueError, json.JSONDecodeError, pydantic.ValidationError):
                        raise ValueError(
                            f"The JWKS endpoint of {self.issuer} returned an unexpected "
                            f"response (url: {response.url})"
                        )

        # RFC 8414: The "issuer" value returned MUST be identical to the authorization
        # server's issuer identifier value into which the well-known URI string was
        # inserted to create the URL used to retrieve the metadata. If these values
        # are not identical, the data contained in the response MUST NOT be used.
        if self.issuer != metadata.issuer:
            raise ValueError(
                f"Metadata discovered from {self.issuer} specifies a different issuer"
                f": {metadata.issuer}"
            )

        for field in ServerMetadata.model_fields:
            if field in self.model_fields_set and not getattr(self, field) is None:
                continue
            setattr(self, field, getattr(metadata, field))

        self.discovered = True
        if metadata.signed_metadata:
            logger.warning(
                "Received unsupported signed_metadata from %s",
                self.issuer
            )

        DISCOVERED[self.issuer] = self
        return True

    @contextlib.asynccontextmanager
    async def _client(self, client: httpx.AsyncClient | None):
        if client is not None:
            yield client
            return
        async with httpx.AsyncClient(base_url=self.issuer, verify=not os.getenv('AEGISX_SSL_VERIFY') == '0') as client:
            yield client

    async def _discover_metadata_endpoint(
        self,
        client: httpx.AsyncClient,
        metadata_url: str | None = None,
        timeout: float = 10.0
    ) -> httpx.Response | None:
        metadata_url = metadata_url or self.metadata_url
        match bool(self.metadata_url):
            case False:
                paths: list[str] = [
                    '.well-known/oauth-authorization-server',
                    '.well-known/openid-configuration'
                ]
                for path in paths:
                    response = await client.get(
                        url=f'{str.strip(self.issuer, '/')}/{path}',
                        follow_redirects=True,
                        timeout=timeout
                    )
                    if response.status_code != 200:
                        continue
                    if str(response.url) != self.metadata_url:
                        self.metadata_url = str(response.url)
                    break
                else:
                    response = None
            case True:
                assert self.metadata_url is not None
                response = await client.get(
                    url=self.metadata_url,
                    follow_redirects=True
                )
                if response.status_code != 200:
                    response = None
        return response


    # TODO: These are non-standard properties.
    #token_endpoint_auth_encryption_alg_values_supported: list[str] | None = []
    #token_endpoint_auth_encryption_enc_values_supported: list[str] | None = []
    #assertion_signing_alg_values_supported: list[str] | None = []
    #assertion_encryption_alg_values_supported: list[str] | None = []
    #assertion_encryption_enc_values_supported: list[str] | None = []
    #required_encrypted_token_endpoint_auth: bool | None = False
    #require_encrypted_assertion: bool | None = False


from ._fields import IDTokenField
from ._codetokenauthorizationresponse import CodeTokenAuthorizationResponse

class CodeIDTokenTokenAuthorizationResponse(CodeTokenAuthorizationResponse):
    id_token: IDTokenField


import functools
from typing import Any
from typing import Callable

import fastapi

from ._requesthandler import RequestHandler


class ResourceRouter(fastapi.APIRouter):
    request_handler_class: type[RequestHandler] = RequestHandler

    def scope(self, scope: str):
        def decorator(func: Callable[..., Any] | RequestHandler) -> RequestHandler:
            if not isinstance(func, self.request_handler_class):
                func = functools.wraps(func)(self.request_handler_class(func))
            return func.with_scope(scope) # type: ignore
        return decorator


from ._accesstokentype import AccessTokenType
from ._authenticationmethodreferenceliteral import AuthenticationMethodReferenceLiteral
from ._clientauthenticationmethod import ClientAuthenticationMethod
from ._granttypeliteral import GrantTypeLiteral
from ._httpsresourcelocator import HTTPSResourceLocator
from ._issueridentifier import IssuerIdentifier
from ._oidcvalidationcontext import OIDCValidationContext
from ._openauthorizationerror import ClientConfigurationError
from ._openauthorizationerror import Error
from ._openauthorizationerror import MetadataError
from ._openauthorizationerror import NeedsDiscovery
from ._openauthorizationerror import NotDiscoverable
from ._responsetype import ResponseType
from ._validationcontext import ValidationContext


__all__: list[str] = [
    'AccessTokenType',
    'AuthenticationMethodReferenceLiteral',
    'ClientAuthenticationMethod',
    'ClientConfigurationError',
    'Error',
    'GrantTypeLiteral',
    'HTTPSResourceLocator',
    'IssuerIdentifier',
    'MetadataError',
    'NeedsDiscovery',
    'NotDiscoverable',
    'OIDCValidationContext',
    'ResponseType',
    'ValidationContext',
]


from ._authorizationresponsebase import AuthorizationResponseBase
from ._fields import IDTokenField


class IDTokenAuthorizationResponse(AuthorizationResponseBase):
    id_stoken: IDTokenField

    def is_encrypted(self) -> bool:
        return False

    def is_signed(self) -> bool:
        return False


import sys
from typing import NoReturn


class OpenAuthorizationError(Exception):
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    default_description: str | None = None

    def __init__(self, description: str | None = None):
        self.description = description or self.default_description

    def fatal(self) -> NoReturn:
        sys.stderr.write(f'{self.FAIL}{self.description}{self.ENDC}')
        raise SystemExit()


class ServerError(OpenAuthorizationError):
    pass


class MetadataError(ServerError):
    pass



class NotDiscoverable(MetadataError):
    pass


class NeedsDiscovery(MetadataError):
    default_description= (
        "The authorization servers' metadata is not discovered yet. Call "
        "ServerMetadata.discover() before using this instance."
    )


class ClientConfigurationError(OpenAuthorizationError):
    pass


class Error(OpenAuthorizationError):
    pass


from ._codeauthorizationresponse import CodeAuthorizationResponse
from ._fields import AccessTokenExpiresInField
from ._fields import AccessTokenField
from ._fields import AccessTokenScopeField
from ._fields import AccessTokenTypeField
from ._fields import StateField
from ._tokenresponsemixin import TokenResponseMixin


class CodeTokenAuthorizationResponse(CodeAuthorizationResponse, TokenResponseMixin):
    access_token: AccessTokenField
    token_type: AccessTokenTypeField
    expires_in: AccessTokenExpiresInField
    scope: AccessTokenScopeField
    state: StateField


from typing import Literal

import pydantic

from ._tokenrequestbase import TokenRequestBase


class RefreshTokenRequest(TokenRequestBase[Literal['refresh_token']]):
    refresh_token: str = pydantic.Field(
        default=...,
        title="Refresh token",
        description="The refresh token issued to the client."
    )


from .client import Client
from .client import ClientSecretCredential
from .client import InteractiveAuth
from .models import ServerMetadata


__all__: list[str] = [
    'Client',
    'ClientSecretCredential',
    'InteractiveAuth',
    'ServerMetadata',
]


from ._baseresourceserver import BaseResourceServerAuth
from ._clientsecretcredential import ClientSecretCredential
from ._interactive import InteractiveAuth


__all__: list[str] = [
    'BaseResourceServerAuth',
    'ClientSecretCredential',
    'InteractiveAuth'
]


from typing import Any
from typing import NoReturn
from typing import TypeVar

import httpx
import pydantic

from aegisx.ext.oauth.types import Error
from ._authorizationserverresponse import AuthorizationServerResponse


E = TypeVar('E', default=str)


class NonProtocolErrorResponse(AuthorizationServerResponse):
    model_config = {'extra': 'allow'}

    @pydantic.model_validator(mode='before')
    @classmethod
    def preprocess(cls, values: Any, info: pydantic.ValidationInfo):
        # Note that this must be after ErrorResponse in a
        # response union.
        if info.context and isinstance(info.context, httpx.Response):
            if info.context.status_code < 400:
                raise ValueError(f"not an error response: {info.context.status_code}")
        return values

    def fatal(self) -> NoReturn:
        Error(repr(self)).fatal()

    def is_error(self):
        return True

    def throw(self) -> NoReturn:
        raise Error(repr(self))


import inspect

import fastapi
from aegisx.ext.iam import AuthorizationContext
from aegisx.ext.iam import AnonymousSubject


class NullRequestContext:

    def __init__(self):
        inspect.markcoroutinefunction(self)
        assert inspect.iscoroutinefunction(self)

    async def __call__(self, request: fastapi.Request):
        assert request.client
        return AuthorizationContext.model_validate({
            'subject': AnonymousSubject(),
            'remote_host': request.client.host
        })


from typing import Annotated

import fastapi


async def get(request: fastapi.Request) -> set[str]:
    return getattr(request.state, 'scope', set())


CurrentScope = Annotated[
    set[str],
    fastapi.Depends(get)
]


import inspect
from typing import cast
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Literal

import fastapi
import fastapi.security
import pydantic
from aegisx.ext.iam import AnonymousSubject
from aegisx.ext.iam import AuthenticatedSubject
from aegisx.ext.jose.types import JWECompactEncoded
from aegisx.ext.jose.types import JWSCompactEncoded

from aegisx.ext.oauth.models import OIDCToken
from ._oidctokenvalidator import OIDCTokenValidator


Subject = AnonymousSubject | AuthenticatedSubject

SubjectFactory = Callable[[OIDCToken], Subject | None | Awaitable[Subject | None]]


class OIDCTokenBearer(fastapi.security.HTTPBearer):
    adapter: pydantic.TypeAdapter[JWECompactEncoded | JWSCompactEncoded]
    audience_mode: Literal['domain', 'path']
    max_age: int
    subjects: set[str]

    def __init__(
        self,
        issuers: set[str],
        audience: set[str] | None = None,
        audience_mode: Literal['domain', 'path'] = 'domain',
        max_age: int = 0,
        subjects: set[str] | None = None,
        subject_factory: SubjectFactory | None = None
    ):
        super().__init__(
            auto_error=False,
            description=("OIDC ID Token")
        )
        self.adapter = pydantic.TypeAdapter(JWECompactEncoded | JWSCompactEncoded)
        self.audience = audience or set()
        self.audience_mode = audience_mode
        self.issuers = issuers
        self.max_age = max_age
        self.subjects = subjects or set()
        self.subject_factory = subject_factory or cast(SubjectFactory, lambda _: AnonymousSubject()) # type: ignore

    def get_audience(self, request: fastapi.Request) -> set[str]:
        return {*self.audience, *self.request_audience(request)}

    def is_service_account(self, token: OIDCToken) -> bool:
        return False

    def request_audience(self, request: fastapi.Request) -> set[str]:
        audience: set[str] = set()
        match self.audience_mode:
            case 'domain':
                audience.add(f'{request.url.scheme}://{request.url.netloc}')
            case 'path':
                audience.add(f'{request.url.scheme}://{request.url.netloc}{request.url.path}')
        return audience

    async def authenticate(
        self,
        request: fastapi.Request,
        bearer: fastapi.security.HTTPAuthorizationCredentials | None
    ):
        request.state.subject = AnonymousSubject()
        if bearer is not None:
            request.state.subject = await self.get_subject(
                await OIDCTokenValidator()
                    .max_age(self.max_age)
                    .with_audiences(self.get_audience(request))
                    .with_issuers(self.issuers)
                    .with_subjects(self.subjects)
                    .validate(
                        self.adapter.validate_python(bearer.credentials)
                    )
            )

    async def get_subject(
        self,
        token: OIDCToken
    ) -> AuthenticatedSubject[Any] | AnonymousSubject:
        subject = self.subject_factory(token)
        if inspect.isawaitable(subject):
            subject = await subject
        return subject or AnonymousSubject()

    async def __call__(self, request: fastapi.Request):
        request.state.subject = AnonymousSubject()
        bearer = await super().__call__(request)
        await self.authenticate(request, bearer)
        return bearer


from ._authorizationrequestparameters import AuthorizationRequestParameters
from ._authorizationresponse import AuthorizationResponse
from ._clientconfiguration import ClientConfiguration
from ._grant import Grant
from ._oidcclaimrequest import ClaimRequest
from ._oidcclaimspec import OIDCClaimSpec
from ._oidctoken import OIDCToken
from ._oidcrequestedclaims import RequestedClaims
from ._servermetadata import ServerMetadata
from ._tokenrequest import TokenRequest
from ._tokenresponse import TokenResponse


__all__: list[str] = [
    'AuthorizationRequestParameters',
    'AuthorizationResponse',
    'ClientConfiguration',
    'ClaimRequest',
    'Grant',
    'OIDCClaimSpec',
    'OIDCToken',
    'RequestedClaims',
    'ServerMetadata',
    'TokenRequest',
    'TokenResponse',
]


import asyncio
import asyncio.coroutines
import inspect

import fastapi


class AuthenticationBackend:
    """The base class for resource server authentication backends."""
    _is_coroutine = asyncio.coroutines._is_coroutine # type: ignore
    priority: int = 1000

    @property
    def __call__(self):
        return self.authenticate

    @property
    def __signature__(self):
        return inspect.signature(self.authenticate)

    def __init__(self):
        # Check if the asyncio.iscoroutinefunction() call returns
        # True for this object, since it depends on a private
        # symbol.
        assert asyncio.iscoroutinefunction(self)

    def can_authenticate(self, request: fastapi.Request) -> bool:
        """Return ``True`` if the backend can authenticate the request."""
        raise NotImplementedError

    def is_authenticated(self, request: fastapi.Request) -> bool:
        return getattr(request.state, 'authenticated', False)

    def handle(self, request: fastapi.Request):
        if self.is_authenticated(request):
            return
        return self.authenticate(request)

    async def authenticate(self, request: fastapi.Request) -> None:
        """Authenticate the request and set the ``principal`` attribute
        on the state. Other claims may be set.
        """
        raise NotImplementedError


class NullAuthenticationBackend(AuthenticationBackend):

    async def authenticate(self, request: fastapi.Request) -> None:
        return


from ._invalidrequestscope import InvalidRequestScope
from ._invalidrequesttoken import InvalidRequestToken
from ._resourceserverexception import ResourceServerException


__all__: list[str] = [
    'InvalidRequestScope',
    'InvalidRequestToken',
    'ResourceServerException',
]


from ._accesstokenbearer import AccessTokenBearer
from ._nullrequestcontext import NullRequestContext
from ._requestcontext import RequestContext

__all__: list[str] = [
    'AccessTokenBearer',
    'NullRequestContext',
    'RequestContext'
]


import pydantic


class AuthorizationServerResponse(pydantic.BaseModel):
    model_config = {'extra': 'forbid'}

    def is_encrypted(self) -> bool:
        raise NotImplementedError

    def is_signed(self) -> bool:
        raise NotImplementedError

    def is_error(self) -> bool:
        return False


from typing import TypedDict
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aegisx.ext.oauth.models import AuthorizationRequestParameters
    from aegisx.ext.oauth.models import ServerMetadata
    from aegisx.ext.oauth.models import TokenResponse
    from aegisx.ext.jose import JSONWebSignature


class OIDCValidationContext(TypedDict):
    client_id: str
    grant: 'TokenResponse'
    jws: 'JSONWebSignature'
    metadata: 'ServerMetadata'
    now: int
    params: 'AuthorizationRequestParameters'


import json
import urllib.parse
from typing import cast
from typing import Any

import pydantic

from ._oidcrequestedclaims import RequestedClaims


class OIDCClaimSpec(pydantic.BaseModel):
    userinfo: RequestedClaims | None = pydantic.Field(
        default=None,
        title="UserInfo Endpoint Claims",
        description=(
            "Requests that the listed individual claims be returned from the UserInfo endpoint. "
            "If present, the listed claims are being requested to be added to any claims that "
            "are being requested using scope values. If not present, the claims being requested "
            "from the UserInfo Endpoint are only those requested using scope values."
        )
    )

    id_token: RequestedClaims | None = pydantic.Field(
        default=None,
        title="ID Token Claims",
        description=(
            "Requests that the listed individual claims be returned in the ID Token. If present, "
            "the listed claims are being requested to be added to the default claims in the ID "
            "Token. If not present, the default ID Token claims are requested."
        )
    )

    @pydantic.model_validator(mode='before')
    @classmethod
    def preprocess(
        cls,
        values: dict[str, Any] | str
    ) -> dict[str, Any]:
        if isinstance(values, str):
            values = cast(
                dict[str, Any],
                json.loads(urllib.parse.unquote_plus(values))
            )
        return values

    @pydantic.model_serializer(mode='wrap', when_used='always')
    def serialize(
        self,
        nxt: pydantic.SerializerFunctionWrapHandler,
        info: pydantic.SerializationInfo
    ):
        match info.mode:
            case 'query':
                return urllib.parse.quote_plus(
                    json.dumps(
                        self.model_dump(mode='json'),
                        separators=(',', ':')
                    )
                )
            case _:
                return nxt(self)


import pydantic

from ._authorizationresponsebase import AuthorizationResponseBase


class CodeAuthorizationResponse(AuthorizationResponseBase):
    # TODO: Google adds certain non-standard parameters to the
    # response.
    model_config = {'extra': 'ignore'}

    code: str | None = pydantic.Field(
        default=None,
        title="Code",
        description=(
            "Present if the `code` was in the requested `response_type`. The "
            "authorization code generated by the authorization server. The "
            "authorization code MUST expire shortly after it is issued to "
            "mitigate the risk of leaks.  A maximum authorization code lifetime "
            "of 10 minutes is RECOMMENDED. The client MUST NOT use the authorization "
            "code more than once. If an authorization code is used more than once, "
            "the authorization server denies the request and revokes (when possible) "
            "all tokens previously issued based on that authorization code. "
            "The authorization code is bound to the client identifier and "
            "redirection URI."
        )
    )

    def is_encrypted(self) -> bool:
        return False

    def is_signed(self) -> bool:
        return False


from typing import ClassVar

import pydantic

from aegisx.ext.oauth.types import GrantTypeLiteral
from ._tokenresponse import TokenResponse


class Grant(pydantic.BaseModel):
    __noindex__: ClassVar[set[str]] = {'response'}

    name: str
    grant_type: GrantTypeLiteral
    issuer: str
    obtained: int
    scope: set[str]
    response: TokenResponse

    @property
    def access_token(self):
        return self.response.access_token

    @property
    def expires_in(self):
        return self.response.expires_in

    @property
    def refresh_token(self):
        return self.response.refresh_token

    @property
    def token_type(self):
        return self.response.token_type

    @pydantic.field_serializer('scope')
    def serialize_set_protobuf(
        self,
        value: set[str],
        info: pydantic.FieldSerializationInfo,
    ):
        return list(sorted(self.scope))


from typing import Annotated

import pydantic
from aegisx.types import SpaceSeparatedSet
from aegisx.ext.jose.types import JWECompactEncoded
from aegisx.ext.jose.types import JWSCompactEncoded

from aegisx.ext.oauth.types import AccessTokenType
from aegisx.ext.oauth.types import IssuerIdentifier


AccessTokenField = Annotated[
    str,
    pydantic.Field(
        default=...
    )
]

AccessTokenTypeField = Annotated[
    AccessTokenType,
    pydantic.Field(
        default=...,
        title="Token type",
        description=(
            "The type of the token issued as described in "
            "Section 7.1 of RFC 6749. Value is case insensitive."
        )
    )
]

AccessTokenExpiresInField = Annotated[
    int | None,
    pydantic.Field(
        default=None,
        title="Expires in",
        description=(
            "The lifetime in seconds of the access token.  For example, the "
            "value `3600` denotes that the access token will expire in one "
            "hour from the time the response was generated. If omitted, the "
            "authorization server SHOULD provide the expiration time via other "
            "means or document the default value."
        )
    )
]

AccessTokenScopeField = Annotated[
    SpaceSeparatedSet,
    pydantic.Field(
        default_factory=SpaceSeparatedSet,
        title="Scope",
        description=(
            "May be present, if identical to the scope requested by the client; "
            "otherwise, it MUST be present. The scope of the access token as "
            "described by Section 3.3 of RFC 6749."
        )
    )
]


AuthorizationRequestScopeField = Annotated[
    SpaceSeparatedSet,
    pydantic.Field(
        default_factory=SpaceSeparatedSet,
        title="Scope",
        description="The scope of the access request."
    )
]

IDTokenField = Annotated[
    JWECompactEncoded | JWSCompactEncoded,
    pydantic.Field(
        default=...,
        title="ID Token",
        description=(
            "The OpenID Connect ID Token received from the authorization server. ID Tokens "
            "MUST be signed using JWS and optionally both signed and then encrypted using "
            "JWS and JWE respectively, thereby providing authentication, integrity, non-"
            "repudiation, and optionally, confidentiality, per Section 16.14 of OpenID "
            "Connect Core 1.0 incorporating errata set 2. If the ID Token is encrypted, "
            "it MUST be signed then encrypted, with the result being a Nested JWT, as "
            "defined in RFC 7519. ID Tokens MUST NOT use none as the `alg` value unless the "
            "Response Type used returns no ID Token from the Authorization Endpoint "
            "(such as when using the Authorization Code Flow) and the Client explicitly "
            "requested the use of none at Registration time."
        )
    )
]

StateField = Annotated[
    str | None,
    pydantic.Field(
        default=None,
        title="State",
        description=(
            "Present if the `state` parameter was present in the client "
            "authorization request.  The exact value received from the "
            "client."
        )
    )
]

# RFC 9702 Section 2: The iss parameter value is the issuer identifier of the
# authorization server that created the authorization response, as defined in
# [RFC8414]. Its value MUST be a URL that uses the "https" scheme without any
# query or fragment components.
IssuerIdentifierField = Annotated[
    IssuerIdentifier | None,
    pydantic.Field(
        default=None,
        title="Issuer identifier",
        description=(
            "The `iss` parameter value is the issuer identifier of the "
            "authorization server that created the authorization response, "
            "as defined in RFC 8414. Its value MUST be a URL that uses the "
            "`https` scheme without any query or fragment components."
        )
    )
]


from typing import Generic
from typing import TypeVar

import pydantic

from aegisx.ext.oauth.types import GrantTypeLiteral


G = TypeVar('G')


class TokenRequestBase(pydantic.BaseModel, Generic[G]):
    grant_type: GrantTypeLiteral = pydantic.Field(
        default=...,
        title="Grant type",
        description=(
            "Specifies the grant type requested by the client."
        )
    )

    client_id: str | None = pydantic.Field(
        default=None,
        title="Client ID",
        description=(
            "Request if the client is not authenticating with the "
            " authorization server as described in Section 3.2.1 of "
            "RFC 6749, or if the client uses the `client_secret_post` "
            "method to authenticate."
        )
    )

    client_secret: str | None = pydantic.Field(
        default=None,
        title="Client secret",
        description=(
            "Request if the client uses the `client_secret_post` "
            "method to authenticate itself."
        )
    )

    @pydantic.model_validator(mode='after')
    def postprocess(self):
        if self.client_secret and not self.client_id:
            raise ValueError(
                "Both the \"client_id\" and \"client_secret\" "
                "parameters must be provided."
            )
        return self